"""sui.dl
Deep learning algorithm implementations
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__api_info_dict = {'nn.PNN': 'Product-based Neural Networks'}


def api_info(api=None):
    if api is not None:
        if api in __api_info_dict:
            return 'API: {}\nInfo: {}\n'.format(api, __api_info_dict[api])
        else:
            return '{} is not an appreciable API under sui.dl'.format(api)
    else:
        for api, info in __api_info_dict.items():
            return 'API: {}\nInfo: {}\n'.format(api, info)



"""Module including some useful implementations apropos neural networks.
"""
__author__ = ['Li Tang']
__copyright__ = 'Li Tang'
__credits__ = ['Li Tang']
__license__ = 'GPL'
__version__ = '1.0.1'
__maintainer__ = ['Li Tang']
__email__ = 'litang1025@gmail.com'
__status__ = 'Production'

import tensorflow as tf
from tensorflow.keras.layers import Dense

from sui import dataset


class PNN(tf.keras.Model):
    """Product-based Neural Networks
    https://arxiv.org/pdf/1611.00144.pdf
    """

    def __init__(self, features_dim: int, fields_dim: int, hidden_layer_sizes: list, dropout_params: list,
                 product_layer_dim=10, lasso=0.01, ridge=1e-5, embedding_dim=10, product_type='ipnn', model=None):
        super(PNN, self).__init__()
        self.features_dim = features_dim  # number of different features after one-hot, denoted by F
        self.fields_dim = fields_dim  # number of different original features
        self.dropout_params = dropout_params
        self.hidden_layer_sizes = hidden_layer_sizes  # number of hidden layers
        self.product_layer_dim = product_layer_dim
        self.lasso = lasso
        self.ridge = ridge
        self.embedding_dim = embedding_dim  # dimension of vectors after embedding, denoted by M
        self.product_type = product_type  # 'ipnn' for inner product while 'opnn' for outer product

        self.model = model
        # if there is no pre-trained model to use
        if self.model is None:
            # embedding layer
            self.embedding_layer = self.__init_embedding_layer()

            # product layer
            self.product_layer = self.__init_product_layer()

            # hidden layers
            for layer_index in range(len(self.hidden_layer_sizes)):
                setattr(self, 'dense_' + str(layer_index), tf.keras.layers.Dense(self.hidden_layer_sizes[layer_index]))
                setattr(self, 'batch_norm_' + str(layer_index), tf.keras.layers.BatchNormalization())
                setattr(self, 'activation_' + str(layer_index), tf.keras.layers.Activation('relu'))
                setattr(self, 'dropout_' + str(layer_index), tf.keras.layers.Dropout(dropout_params[layer_index]))

    def __init_embedding_layer(self):
        # the size of embedding layer is F * M
        return tf.keras.layers.Embedding(self.features_dim, self.embedding_dim, embeddings_initializer='uniform')

    def __init_product_layer(self):
        # linear signals l_z
        self.__linear_weights = self.__init_linear_signals()
        # embedding_feature = tf.einsum('bnm,bn->bnm', self.embedding_layer(input_feature_index), input_feature_value)
        # l_z = tf.einsum('bnm,dnm->bd', embedding_feature, self.__linear_weights)
        l_z = self.embedding_layer

        # # quadratic signals l_p
        self.__quadratic_signals_variable = self.__init_quadratic_signals()
        # l_p = tf.einsum('bdnm,bdnm->bd', self.__quadratic_signals_variable,
        #                 self.__quadratic_signals_variable)  # Batch * D1
        l_p = self.embedding_layer
        # return tf.concat((l_z, l_p), axis=1)
        return tf.keras.layers.Dense(self.hidden_layer_sizes[0])

    def __init_linear_signals(self, initializer=tf.initializers.GlorotUniform()):
        return tf.Variable(initializer(shape=(self.product_layer_dim, self.fields_dim, self.embedding_dim)))

    def __init_quadratic_signals(self, initializer=tf.initializers.GlorotUniform()):
        assert self.product_type in ['ipnn', 'opnn'], "'product_type' should be either 'ipnn' or 'opnn'."
        if self.product_type == 'ipnn':
            # matrix decomposition based on the assumption: W_p^n = \theta ^n * {\theta^n}^T
            return tf.Variable(initializer(shape=(self.product_layer_dim, self.fields_dim)))
        elif self.product_type == 'opnn':
            pass
        else:
            raise Exception('Arcane exception...')

    @staticmethod
    def loss_function(labels, logits, name=None):
        return tf.nn.sigmoid_cross_entropy_with_logits(labels=labels, logits=logits, name=name)

    def __create_model(self, training=False):
        model = tf.keras.Sequential()

        # embedding layer
        model.add(self.embedding_layer)

        # product layer
        model.add(self.product_layer)

        # fc layers
        for layer_index in range(len(self.hidden_layer_sizes)):
            model.add(getattr(self, 'dense_' + str(layer_index)))
            model.add(getattr(self, 'batch_norm_' + str(layer_index)))
            model.add(getattr(self, 'activation_' + str(layer_index)))
            if training:
                model.add(getattr(pnn, 'dropout_' + str(layer_index)))

        # output layer
        model.add(Dense(2, activation='sigmoid', kernel_initializer='he_normal', name='output'))

        # compile
        model.compile(loss='sparse_categorical_crossentropy', optimizer='adam', metrics=['accuracy'])

        self.model = model

    def train(self, data, labels, batch_size=None, epochs=1, verbose=1):
        if self.model is None:
            self.__create_model(training=True)
        self.model.fit(x=data, y=labels, batch_size=batch_size, epochs=epochs, verbose=verbose)

    def predict(self, data, batch_size=None, verbose=0, steps=None, callbacks=None, max_queue_size=10, workers=1,
                use_multiprocessing=False):
        return self.model.predict(x=data, batch_size=batch_size, verbose=verbose, steps=steps, callbacks=callbacks,
                                  max_queue_size=max_queue_size, workers=workers,
                                  use_multiprocessing=use_multiprocessing)

    # TODO dump
    def dump(self):
        pass

    def restore(self):
        pass



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__author__ = ['Li Tang']
__copyright__ = 'Li Tang'
__credits__ = ['Li Tang']
__license__ = 'MIT'
__version__ = '0.1.6'
__maintainer__ = ['Li Tang']
__email__ = 'litang1025@gmail.com'
__status__ = 'Production'

my_dear = "Dear Miss Sui Lin, I love you!"



import tensorflow as tf
from python import utils

class Model(tf.keras.Model):
    def __init__(self):
        super(Model, self).__init__()


class PNN1(Model):
    def __init__(self, field_sizes=None, embed_size=10, layer_sizes=None, layer_acts=None, drop_out=None,
                 embed_l2=None, layer_l2=None, init_path=None, opt_algo='gd', learning_rate=1e-2, random_seed=None):
        Model.__init__(self)
        init_vars = []
        num_inputs = len(field_sizes)
        for i in range(num_inputs):
            init_vars.append(('embed_%d' % i, [field_sizes[i], embed_size], 'xavier', dtype))
        num_pairs = int(num_inputs * (num_inputs - 1) / 2)
        node_in = num_inputs * embed_size + num_pairs
        # node_in = num_inputs * (embed_size + num_inputs)
        for i in range(len(layer_sizes)):
            init_vars.append(('w%d' % i, [node_in, layer_sizes[i]], 'xavier', dtype))
            init_vars.append(('b%d' % i, [layer_sizes[i]], 'zero', dtype))
            node_in = layer_sizes[i]
        self.graph = tf.Graph()
        with self.graph.as_default():
            if random_seed is not None:
                tf.set_random_seed(random_seed)
            self.X = [tf.sparse_placeholder(dtype) for _ in range(num_inputs)]
            self.y = tf.placeholder(dtype)
            self.keep_prob_train = 1 - np.array(drop_out)
            self.keep_prob_test = np.ones_like(drop_out)
            self.layer_keeps = tf.placeholder(dtype)
            self.vars = utils.init_var_map(init_vars, init_path)
            w0 = [self.vars['embed_%d' % i] for i in range(num_inputs)]
            xw = tf.concat([tf.sparse_tensor_dense_matmul(self.X[i], w0[i]) for i in range(num_inputs)], 1)
            xw3d = tf.reshape(xw, [-1, num_inputs, embed_size])

            row = []
            col = []
            for i in range(num_inputs-1):
                for j in range(i+1, num_inputs):
                    row.append(i)
                    col.append(j)
            # batch * pair * k
            p = tf.transpose(
                # pair * batch * k
                tf.gather(
                    # num * batch * k
                    tf.transpose(
                        xw3d, [1, 0, 2]),
                    row),
                [1, 0, 2])
            # batch * pair * k
            q = tf.transpose(
                tf.gather(
                    tf.transpose(
                        xw3d, [1, 0, 2]),
                    col),
                [1, 0, 2])
            p = tf.reshape(p, [-1, num_pairs, embed_size])
            q = tf.reshape(q, [-1, num_pairs, embed_size])
            ip = tf.reshape(tf.reduce_sum(p * q, [-1]), [-1, num_pairs])

            # simple but redundant
            # batch * n * 1 * k, batch * 1 * n * k
            # ip = tf.reshape(
            #     tf.reduce_sum(
            #         tf.expand_dims(xw3d, 2) *
            #         tf.expand_dims(xw3d, 1),
            #         3),
            #     [-1, num_inputs**2])
            l = tf.concat([xw, ip], 1)

            for i in range(len(layer_sizes)):
                wi = self.vars['w%d' % i]
                bi = self.vars['b%d' % i]
                l = tf.nn.dropout(
                    utils.activate(
                        tf.matmul(l, wi) + bi,
                        layer_acts[i]),
                    self.layer_keeps[i])

            l = tf.squeeze(l)
            self.y_prob = tf.sigmoid(l)

            self.loss = tf.reduce_mean(
                tf.nn.sigmoid_cross_entropy_with_logits(logits=l, labels=self.y))
            if layer_l2 is not None:
                self.loss += embed_l2 * tf.nn.l2_loss(xw)
                for i in range(len(layer_sizes)):
                    wi = self.vars['w%d' % i]
                    self.loss += layer_l2[i] * tf.nn.l2_loss(wi)
            self.optimizer = utils.get_optimizer(opt_algo, learning_rate, self.loss)

            config = tf.ConfigProto()
            config.gpu_options.allow_growth = True
            self.sess = tf.Session(config=config)
            tf.global_variables_initializer().run(session=self.sess)